#include "Aligner.h"
#include "NGM.h"

#undef module_name
#define module_name "ALIGN"

/*
 *  Aligner ist eine Thread/CUDA/GPU-Einheit zur Berechnung von Scores/Alignments
 *
 *
 */
class AlignmentDispatcher::Aligner
{
private:
	NGMMutex m_Mutex;
	NGMThreadWait m_EntryWait;
	NGMThreadWait m_ReturnWait;

	int const m_GPU;
	int const m_OutputFormat;

	IAlignment * m_Kernel;
	volatile bool m_KernelInit;

	bool m_Running;
	bool m_Finished;

	int m_IvkType;

	int m_IvkMode;
	int m_IvkBatchSize;
	char const * const * m_IvkRefSeqList;
	char const * const * m_IvkQrySeqList;
	void * m_IvkResults;
	void * m_IvkData;
	int m_IvkReturn;

	static NGMTHREADFUNC Run(void* pObj)
	{
		Aligner * align = (Aligner*)pObj;

		Log.Message("Alignment Thread for GPU %i startet", align->m_GPU);
		NGMLock(&align->m_Mutex);
		align->m_Kernel = CreateAlignment( align->m_GPU | (align->m_OutputFormat << 8) );
		if (align->m_Kernel == 0)
			Fatal();

		NGMSignal(&align->m_ReturnWait);

		while (true)
		{
			while (align->m_IvkType <= 0)
			{
				if (align->m_Finished)
				{
					NGMUnlock(&align->m_Mutex);
					Log.Message("Alignment Thread for GPU %i finished", align->m_GPU);
					return 0;
				}

				//Log.Green("Waiting");
				NGMWait(&align->m_Mutex, &align->m_EntryWait);
			}
			Log.Green("Launching GPU...");

			if (align->m_IvkType == 1)
				align->m_IvkReturn = align->m_Kernel->BatchScore(align->m_IvkMode, align->m_IvkBatchSize, align->m_IvkRefSeqList, align->m_IvkQrySeqList, (float*)align->m_IvkResults, align->m_IvkData);
			else if (align->m_IvkType == 2)
				align->m_IvkReturn = align->m_Kernel->BatchAlign(align->m_IvkMode, align->m_IvkBatchSize, align->m_IvkRefSeqList, align->m_IvkQrySeqList, (Align*)align->m_IvkResults, align->m_IvkData);
			else
				align->m_IvkReturn = 0;

			align->m_IvkType = 0;

			//Log.Green("done");
			NGMSignal(&align->m_ReturnWait);
			//Log.Green("signalled");

		}
	}
public:
	bool InUse;

	static pfCreateAlignment CreateAlignment;
	static pfDeleteAlignment DeleteAlignment;

	Aligner(int const gpu) :
		m_Mutex(),
		m_EntryWait(),
		m_ReturnWait(),
		m_GPU( gpu ),
		m_OutputFormat( Config.GetInt("outputformat", 0, 1) ),
		m_Kernel( 0 ),
		m_KernelInit( false ),
		m_Running( false ),
		m_Finished( false ),
		m_IvkType( 0 ),
		InUse( false )
	{
		NGMInitMutex(&m_Mutex);
		NGMInitWait(&m_EntryWait);
		NGMInitWait(&m_ReturnWait);

		NGMCreateThread(&Aligner::Run, this, true);
	}

	int Dispatch(int type, int const mode,
			int const batchSize,
			char const * const * const refSeqList,
			char const * const * const qrySeqList,
			void * const results,
			void * extData)
	{
		NGMLock(&m_Mutex);
		if (m_Running)
		{
			NGMUnlock(&m_Mutex);
			return -1;
		}

		m_IvkType = type;
		m_IvkMode = mode  | (m_OutputFormat << 8);
		m_IvkBatchSize = batchSize;
		m_IvkRefSeqList = refSeqList;
		m_IvkQrySeqList = qrySeqList;
		m_IvkResults = results;
		m_IvkData = extData;
		m_IvkReturn = 0;

		m_Running = true;

		NGMSignal(&m_EntryWait);

		while (m_IvkType > 0)
		{
			//Log.Error("Waiting");
			NGMWait(&m_Mutex, &m_ReturnWait);
		}

		int ret = m_IvkReturn;
		m_Running = false;
		NGMUnlock(&m_Mutex);

		return ret;
	}

	IAlignment const * Kernel()
	{
		if (m_Kernel == 0)
		{
			NGMLock(&m_Mutex);
			while (m_Kernel == 0)
				NGMWait(&m_Mutex, &m_ReturnWait);
			NGMUnlock(&m_Mutex);
		}
		return m_Kernel;
	}

	void Shutdown()
	{
		NGMLock(&m_Mutex);
		m_Finished = true;
		NGMSignal(&m_EntryWait);
		NGMUnlock(&m_Mutex);
	}
};

pfCreateAlignment AlignmentDispatcher::Aligner::CreateAlignment = 0;
pfDeleteAlignment AlignmentDispatcher::Aligner::DeleteAlignment = 0;

AlignmentDispatcher * AlignmentDispatcher::sInstance = 0;

NGMOnceControl dispatcher_once = NGM_ONCE_INIT;

void AlignmentDispatcher::CreateInstance()
{
	sInstance = new AlignmentDispatcher();
}

AlignmentDispatcher * AlignmentDispatcher::Instance()
{
	if (sInstance == 0)
	{
		NGMOnce(&dispatcher_once, &AlignmentDispatcher::CreateInstance);
	}
	return sInstance;
}
void AlignmentDispatcher::Shutdown()
{
	for (int i = 0; i < m_TotalAligner; ++i)
	{
		m_Aligner[i]->Shutdown();
		delete m_Aligner[i];
	}
	delete this;
	sInstance = 0;
}

AlignmentDispatcher::AlignmentDispatcher() : m_SelectionMutex(), m_SelectionWait()
{
	NGMInitMutex(&m_SelectionMutex);
	NGMInitWait(&m_SelectionWait);

	m_Aligner[0] = 0;

	m_TotalAligner = m_FreeAligner = InitAligners();
	Log.Message("Dispatcher with %i Aligners created", m_TotalAligner);
}

AlignmentDispatcher::~AlignmentDispatcher()
{
}

int AlignmentDispatcher::InitAligners()
{
	int const dll = InitDLL(Config.GetString("alignment_dll"));
	if (dll == -1)
		Fatal();

	typedef int (*pfCookie)();
	pfCookie pfC = (pfCookie)GetDLLFunc(dll, "Cookie", true);
	if ( pfC() != cCookie )
	{
		Log.Error("Invalid DLL version. Please update and rebuild the DLL from SVN");
		Fatal();
	}

	Aligner::CreateAlignment = (pfCreateAlignment)GetDLLFunc(dll, "CreateAlignment");
	Aligner::DeleteAlignment = (pfDeleteAlignment)GetDLLFunc(dll, "DeleteAlignment", false);

	int const threadcount = Config.GetInt("alignment_threads", 1, cMaxAligner);
	Log.Message("Alignment Threads: %i", threadcount);
	int * gpus = new int[threadcount];
	Config.GetIntArray("alignment_threads", gpus, threadcount);

	for (int i = 0; i < threadcount; ++i)
	{
		m_Aligner[i] = new Aligner(gpus[i]);
	}

	delete[] gpus;
	return threadcount;
}

int AlignmentDispatcher::GetScoreBatchSize() const
{
	//IAlignment const * kernel = m_Aligner[0]->Kernel();
	return m_Aligner[0]->Kernel()->GetScoreBatchSize();
}

int AlignmentDispatcher::GetAlignBatchSize() const
{
//	return 10;
	return m_Aligner[0]->Kernel()->GetScoreBatchSize();
}

int AlignmentDispatcher::BatchScore(
		int const mode,
		int const batchSize,
		char const * const * const refSeqList,
		char const * const * const qrySeqList,
		float * const results,
		void * extData)
{
	Aligner * aligner = PickFreeAligner();
	int res = aligner->Dispatch(1, mode, batchSize, refSeqList, qrySeqList, results, extData);
	aligner->InUse = false; ++m_FreeAligner;
	NGMSignal(&m_SelectionWait);
	return res;
}

int AlignmentDispatcher::BatchAlign(
		int const mode,
		int const batchSize,
		char const * const * const refSeqList,
		char const * const * const qrySeqList,
		Align * const results,
		void * extData)
{
	Aligner * aligner = PickFreeAligner();
	int res = aligner->Dispatch(2, mode, batchSize, refSeqList, qrySeqList, results, extData);
	aligner->InUse = false; ++m_FreeAligner;
	NGMSignal(&m_SelectionWait);
	return res;
}

AlignmentDispatcher::Aligner * AlignmentDispatcher::PickFreeAligner()
{
	Aligner * aligner = 0;
	NGMLock(&m_SelectionMutex);

	while ( !(m_FreeAligner > 0) )
		NGMWait(&m_SelectionMutex, &m_SelectionWait);

	int i = 0;
	while (aligner == 0 && i < m_TotalAligner)
	{
		if (!m_Aligner[i]->InUse)
			aligner = m_Aligner[i];
		else
			++i;
	}
	if (aligner == 0)
		Log.Error("No aligner found");

	aligner->InUse = true;
	--m_FreeAligner;
	NGMUnlock(&m_SelectionMutex);
	return aligner;
}
