#include "NGM.h"

#undef module_name
#define module_name "NGM"

#include <memory.h>
#include <stdlib.h>

#include "Debug.h"

_NGM * _NGM::pInstance = 0;
NGMOnceControl _NGM::once_control = NGM_ONCE_INIT;
char const * _NGM::AppName = 0;

namespace __NGM
{
	inline int min(int a, int b)
	{
		return (a < b) ? a : b;
	}
}

void _NGM::Init()
{
	pInstance = new _NGM();
}

_NGM & _NGM::Instance()
{
	NGMOnce(&_NGM::once_control, Init);
	return *pInstance;
}

_NGM::_NGM() :
		Stats( NGMStats::InitStats(AppName) ),
		bCSSW( "CSSW", 1000000, &Stats->Buffer1 ),
		bSWO( "SWO", 1000000, &Stats->Buffer2 ),
		m_ActiveThreads( 0 ),
		m_NextThread( 0 ),
		m_Paired( Config.GetInt("paired") != 0 ),
		m_OutputFormat( Config.GetInt("outputformat", 0, 1) ),
		m_MappedReads( 0 )
{
	Log.Message("NGM Core initialization");
	NGMInitMutex(&m_Mutex);
	NGMInitMutex(&m_OutputMutex);
	NGMInitMutex(&m_UMRMutex);
	memset(m_StageThreadCount, 0, cMaxStage * sizeof(int));
	m_DualStrand = (Config.GetInt("dualstrand") != 0);

	m_NextRead = Config.GetInt("qry_start") * ((m_Paired) ? 2 : 1);
	m_ReadCount = Config.GetInt("qry_count");
	if (m_ReadCount > 0)
		m_ReadCount *= ((m_Paired) ? 2 : 1);
}
_NGM::~_NGM()
{
}

void _NGM::StartThread( NGMTask * task, int cpu)
{
	Log.Message("Starting thread %i <%s> on cpu %i", m_NextThread, task->GetName(), cpu);
	NGMLock( &m_Mutex );
	task->m_TID = m_NextThread;
	m_Tasks[m_NextThread] = task;
	m_Threads[m_NextThread] = NGMCreateThread(&_NGM::ThreadFunc, task, true);

	if (cpu != -1)
		NGMSetThreadAffinity(&m_Threads[m_NextThread], cpu);

	++m_StageThreadCount[task->GetStage()];
	++m_NextThread;
	++m_ActiveThreads;
	NGMUnlock( &m_Mutex );
}

void _NGM::AddUnmappedRead( int readid, int reason )
{
	NGMLock( &m_UMRMutex );
	m_UnmappedReads.push_back(readid);
	NGMUnlock( &m_UMRMutex );
}

int _NGM::GetUnmappedReadCount() const
{
	try
	{
		return m_UnmappedReads.size();
	}
	catch (...)
	{
		Log.Error("Unable to get unmapped read count");
	}
	return 0;
}

void _NGM::AddMappedRead( int readid )
{
	AtomicInc( &m_MappedReads );
}
int _NGM::GetMappedReadCount() const
{
	return m_MappedReads;
}


int _NGM::GetStageThreadCount( int stage )
{
	NGMLock( &m_Mutex );
	int cnt = 0;
	for (int i = 0; i <= stage; ++i)
	{
		cnt += m_StageThreadCount[i];
	}
	NGMUnlock( &m_Mutex );
	return cnt;
}

void _NGM::FinishStage( int tid )
{
	NGMLock( &m_Mutex );
	int stage = m_Tasks[tid]->GetStage();
	--m_StageThreadCount[stage];
	NGMUnlock( &m_Mutex );
	Log.Message("Thread %i finished its stage (Stage %i, now %i active)", tid, stage, m_StageThreadCount[stage]);
}

void _NGM::FinishThread( int tid )
{
	m_Tasks[tid]->FinishStage();
	NGMLock( &m_Mutex );
	--m_ActiveThreads;
	NGMUnlock( &m_Mutex );
	delete m_Tasks[tid];
	m_Tasks[tid] = 0;
	Log.Message("Thread %i finished (%i worker threads remaining)", tid, m_ActiveThreads);
}

void _NGM::UpdateReadCount(int n)
{
	if (m_NextRead > n || m_NextRead < 0)
	{
		Log.Error("Invalid start read.");
		Fatal();
	}

	if ( (m_ReadCount < 0) || ((m_NextRead + m_ReadCount) > n) )
		m_ReadCount = n - m_NextRead;
}

// Liefert eine Anzahl ReadIds <= desBatchSize, startend bei Read # readId
// oder 0, wenn alle Reads bearbeitet wurden
int const _NGM::GetNextReadBatch(int desBatchSize, int & readId)
{
	NGMLock(&m_Mutex);

	readId = m_NextRead;
	if (m_NextRead != -1)
	{
		if (desBatchSize > m_ReadCount)
		{
			desBatchSize = m_ReadCount;
			m_NextRead = -1;
			Log.Message("Last batch");
		}
		else
			m_NextRead += desBatchSize;

		m_ReadCount -= desBatchSize;
	}
	else
	{
		desBatchSize = 0;
	}
	NGMUnlock(&m_Mutex);
	return desBatchSize;
}

NGMTHREADFUNC _NGM::ThreadFunc(void* data)
{
	NGMTask * task = (NGMTask*)data;
	int tid = task->m_TID;

	try
	{
		task->Run();
		NGM.FinishThread(tid);
	}
	catch (...)
	{
		Log.Error("Unhandled exception in thread %i", tid);
	}

	Log.Message("ThreadFunc on thread %i returning", tid);

	return 0;
}

void _NGM::InitQuit()
{
	static int quitState = 0;
	++quitState;
	if (quitState == 1)
	{
		NGMLock(&m_Mutex);
		m_NextRead = -1;
		NGMUnlock(&m_Mutex);
		Log.Warning("Initiated shutdown - waiting for threads to finish");
	}
	else if (quitState >= 3)
	{
		CleanupPlatform();
		Log.Error("Terminate by user request");
		Log.Message("%i Threads still active", m_ActiveThreads);
		for (int i = 0; i < cMaxStage; ++i)
		{
			if (m_StageThreadCount[i] > 0)
				Log.Message("Stage %i got %i threads still running", i, m_StageThreadCount[i]);
		}
		exit(0);
	}
}

void _NGM::AquireOutputLock()
{
	NGMLock(&m_OutputMutex);
}
void _NGM::ReleaseOutputLock()
{
	NGMUnlock(&m_OutputMutex);
}
